# Copyright (C) 2017 O.S. Systems Software LTDA.
# This software is released under the MIT License

from datetime import datetime, timedelta

from jose import jwt as jose_jwt

from . import oauth2
from .clients import ClientManager
from .jwk import JWK, JWKManager


class Hydra(oauth2.Client):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.clients = ClientManager(self)
        self.jwk = JWKManager(self)

    def challenge_decode(self, jwt):
        jwk = self.jwk.get('hydra.consent.challenge', JWK.PUBLIC)
        key = {'keys': [jwk.jwk]}  # hack to bypass python-jose
        options = {'verify_aud': False}
        return jose_jwt.decode(jwt, key, algorithms='RS256', options=options)

    def challenge_encode(self, challenge, sub, **kwargs):
        exp = int((datetime.utcnow() + timedelta(minutes=10)).timestamp())
        claims = {
            'sub': sub,
            'exp': exp,
            'iat': int(datetime.utcnow().timestamp()),
            'jti': challenge['jti'],
            'aud': challenge['aud'],
            'scp': challenge['scp'],
        }
        claims.update(kwargs)
        jwk = self.jwk.get('hydra.consent.response', JWK.PRIVATE)
        return jose_jwt.encode(claims, jwk.to_rsa(), algorithm='RS256')



# Copyright (C) 2017 O.S. Systems Software LTDA.
# This software is released under the MIT License

from datetime import datetime, timedelta
from urllib.parse import urljoin

import requests


class Token:

    def __init__(self, **kwargs):
        self.issue_time = datetime.now()
        self.expires_in = kwargs.get('expires_in', 0)
        self.scope = kwargs.get('scope')
        self.token = kwargs.get('access_token')
        self.type = kwargs.get('token_type')

    def is_expired(self):
        expiration_date = self.issue_time + timedelta(seconds=self.expires_in)
        return datetime.now() > expiration_date

    def __str__(self):
        return '{} {}'.format(self.type, self.token)


class Client:

    def __init__(self, host, client, secret):
        self.host = host
        self.client = client
        self.secret = secret
        self._tokens = {}

    def request(self, method, path, token=True, **kwargs):
        url = urljoin(self.host, path)
        if token:
            return self._token_request(method, url, **kwargs)
        return self._basic_request(method, url, **kwargs)

    def _basic_request(self, method, url, **kwargs):
        kwargs['auth'] = (self.client, self.secret)
        return requests.request(method, url, **kwargs)

    def _token_request(self, method, url, scope=None, **kwargs):
        token = self.get_access_token(scope)
        headers = kwargs.setdefault('headers', {})
        headers['Authorization'] = str(token)
        return requests.request(method, url, **kwargs)

    def get_access_token(self, scope=None):
        token = self._tokens.get(scope)
        if token is not None and not token.is_expired():
            return token
        data = {'grant_type': 'client_credentials'}

        if scope is not None:
            data['scope'] = scope
        response = self.request(
            'POST', '/oauth2/token', token=False, data=data)
        if response.ok:
            token = Token(**response.json())
            self._tokens[scope] = token
            return token

    def instrospect_token(self, token):
        response = self.request(
            'POST', '/oauth2/introspect', data={'token': token.token})
        if response.ok:
            return response.json()

    def revoke_token(self, token):
        response = self.request(
            'POST', '/oauth2/revoke', token=False, data={'token': token.token})
        return response.ok



# Copyright (C) 2017 O.S. Systems Software LTDA.
# This software is released under the MIT License

from .base import HydraManager


class Client:

    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.owner = kwargs.get('owner')
        self.name = kwargs.get('name') or kwargs.get('client_name')
        self.secret = kwargs.get('secret') or kwargs.get('client_secret')
        self.uri = kwargs.get('uri') or kwargs.get('client_uri')
        self.policy_uri = kwargs.get('policy_uri')
        self.tos_uri = kwargs.get('tos_uri')
        self.logo_uri = kwargs.get('logo_uri')
        self.contacts = kwargs.get('contacts')
        self.redirect_uris = kwargs.get('redirect_uris')
        self.grant_types = kwargs.get('grant_types')
        self.response_types = kwargs.get('response_types')
        self.is_public = kwargs.get('public')
        self.scopes = kwargs.get('scope', '').split() or kwargs.get('scopes', [])  # nopep8

    def as_dict(self):
        data = {
            'id': self.id,
            'owner': self.owner,
            'client_name': self.name,
            'client_secret': self.secret,
            'client_uri': self.uri,
            'policy_uri': self.policy_uri,
            'tos_uri': self.tos_uri,
            'logo_uri': self.logo_uri,
            'contacts': self.contacts,
            'scope': ' '.join(self.scopes),
            'redirect_uris': self.redirect_uris,
            'grant_types': self.grant_types,
            'response_types': self.response_types,
            'public': self.is_public,
        }
        return {k: v for k, v in data.items() if v is not None}


class ClientManager(HydraManager):

    SCOPE = 'hydra.clients'

    def create(self, client):
        response = self.hydra.request(
            'POST', '/clients', scope=self.SCOPE, json=client.as_dict())
        if response.ok:
            return Client(**response.json())

    def get(self, client_id):
        path = '/clients/{}'.format(client_id)
        response = self.hydra.request('GET', path, scope=self.SCOPE)
        if response.ok:
            return Client(**response.json())

    def update(self, client):
        path = '/clients/{}'.format(client.id)
        response = self.hydra.request(
            'PUT', path, scope=self.SCOPE, json=client.as_dict())
        if response.ok:
            return Client(**response.json())

    def delete(self, client_id):
        path = '/clients/{}'.format(client_id)
        self.hydra.request('DELETE', path, scope=self.SCOPE)

    def all(self):
        response = self.hydra.request('GET', '/clients', scope=self.SCOPE)
        if response.ok:
            return [Client(**data) for data in response.json().values()]



# Copyright (C) 2017 O.S. Systems Software LTDA.
# This software is released under the MIT License

import base64
import json

from Crypto.PublicKey import RSA

from .base import HydraManager
from .exceptions import HydraResponseError, HydraRequestError


class JWK:

    PUBLIC = 'public'
    PRIVATE = 'private'

    def __init__(self, **jwk):
        self.jwk = jwk
        self.kty = jwk.get('kty')
        self.alg = jwk.get('alg')
        self.kid = jwk.get('kid')

        self.n = jwk.get('n')
        self.e = jwk.get('e')
        self.d = jwk.get('d')

        self.type = None
        if all([self.n, self.e, self.d]):
            self.type = self.PRIVATE
        elif all([self.n, self.e]):
            self.type = self.PUBLIC

    @staticmethod
    def b64_to_int(data):
        return int.from_bytes(base64.urlsafe_b64decode(data + '=='), 'big')

    def to_rsa(self):
        """Converts a JWK to RSA."""
        numbers = [self.n, self.e]
        if self.type == self.PRIVATE:
            numbers.append(self.d)
        return RSA.construct(tuple(self.b64_to_int(n) for n in numbers))


class JWKManager(HydraManager):

    def get(self, key, type_):
        path = '/keys/{}/{}'.format(key, type_)
        response = self.hydra.request('GET', path, scope='hydra.keys.get')
        if not response.ok:
            raise HydraRequestError('Request to get JWK failed.')
        try:
            jwk = response.json()['keys'][0]
        except json.decoder.JSONDecodeError:
            raise HydraResponseError(
                'Bad response content type (expected JSON)')
        except KeyError:
            raise HydraResponseError(
                'Bad response content (expected to have a keys key)')
        except IndexError:
            raise HydraResponseError('Bad response (empty key set)')
        else:
            return JWK(**jwk)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# This software is released under the MIT License

from setuptools import setup


setup(
    name='hydra-sdk',
    description='Go Hydra SDK for Python',
    keywords='hydra oauth2 openid go',
    version='0.9.1',
    packages=['hydra'],
    install_requires=[
        'pycryptodome==3.6.6',
        'python-jose==3.0.1',
        'requests==2.19.1',
    ],
    author='O.S. Systems Software LTDA',
    author_email='contato@ossystems.com.br',
    url='http://www.ossystems.com.br',
    license='MIT',
    zip_safe=True,
)
